import torch
import torch.nn as nn
import torch.nn.functional as F
from .config import OrpheusConfig

class RMSNorm(nn.Module):
    def __init__(self, hidden_size, eps=1e-6):
        super().__init__()
        self.weight = nn.Parameter(torch.full((hidden_size,), 0.4))
        self.variance_epsilon = eps

    def forward(self, hidden_states):
        input_dtype = hidden_states.dtype
        hidden_states = hidden_states.to(torch.float32)
        variance = hidden_states.pow(2).mean(-1, keepdim=True)
        hidden_states = hidden_states * torch.rsqrt(variance + self.variance_epsilon)
        return self.weight * hidden_states.to(input_dtype)

class SwiGLU(nn.Module):
    def forward(self, x):
        x, gate = x.chunk(2, dim=-1)
        return F.silu(gate) * x

class ProjectionLayer(nn.Module):
    def __init__(self, stack_factor: int = 8):
        super().__init__()
        self.stack_factor = stack_factor

    def _pad_and_stack(self, audio_embeds: torch.Tensor) -> torch.Tensor:
        B, T, C = audio_embeds.shape
        audio_embeds = F.pad(
            audio_embeds, (0, 0, 0, self.stack_factor - T % self.stack_factor)
        )
        B, T, C = audio_embeds.shape
        audio_embeds = audio_embeds.view(
            B, T // self.stack_factor, C * self.stack_factor
        )
        return audio_embeds
    

class OrpheusProjector(ProjectionLayer):
    def __init__(self, config: OrpheusConfig):
        self.hidden_dim = config.hidden_size
        super().__init__(config.stack_factor)
        self.ln_pre = RMSNorm(config.audio_hidden_size * self.stack_factor)
        self.linear_1 = nn.Linear(
            config.audio_hidden_size * self.stack_factor,
            self.hidden_dim,
            bias=False,
        )
        self.act = SwiGLU()
        self.linear_2 = nn.Linear(
            self.hidden_dim // 2, self.hidden_dim, bias=False
        )
        self.ln_post = RMSNorm(self.hidden_dim)

    def forward(self, audio_features: torch.Tensor) -> torch.Tensor:
        audio_features = self._pad_and_stack(audio_features)
        audio_features = self.ln_pre(audio_features)
        hidden_states = self.linear_1(audio_features)
        hidden_states = self.act(hidden_states)
        hidden_states = self.linear_2(hidden_states)
        hidden_states = self.ln_post(hidden_states)
        return hidden_states



from typing import List, Optional, Tuple, Union
import torch
from torch import nn
from torch.nn import functional as F
from transformers import (
    AutoModel,
    PreTrainedModel,
    AutoModelForCausalLM, 
    AutoConfig
)
from dataclasses import dataclass
from transformers.cache_utils import Cache
from transformers.modeling_outputs import ModelOutput





from .components import OrpheusProjector
from .config import OrpheusConfig

@dataclass
class OrpheusCausalLMOutputWithPast(ModelOutput):
    loss: Optional[torch.FloatTensor] = None
    logits: torch.FloatTensor = None
    past_key_values: Optional[List[torch.FloatTensor]] = None
    hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    attentions: Optional[Tuple[torch.FloatTensor]] = None
    audio_hidden_states: Optional[Tuple[torch.FloatTensor]] = None

class OrpheusPreTrainedModel(PreTrainedModel):
    config_class = OrpheusConfig
    supports_gradient_checkpointing = True
    _skip_keys_device_placement = "past_key_values"
    _supports_flash_attn_2 = True

    def _init_weights(self, module):
        std = (
            self.config.initializer_range
            if hasattr(self.config, "initializer_range")
            else self.config.text_config.initializer_range
        )

        if hasattr(module, "class_embedding"):
            module.class_embedding.data.normal_(mean=0.0, std=std)

        if isinstance(module, (nn.Linear, nn.Conv2d)):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.Embedding):
            module.weight.data.normal_(mean=0.0, std=std)
            if module.padding_idx is not None:
                module.weight.data[module.padding_idx].zero_()

class OrpheusForConditionalGeneration(OrpheusPreTrainedModel):
    def __init__(self, config: OrpheusConfig, new_vocab_size=False):
        super().__init__(config)

        self.multi_modal_projector = OrpheusProjector(config)

        self.vocab_size = config.vocab_size
        if config.text_model_id is not None:
            self.language_model = AutoModelForCausalLM.from_pretrained(
                config.text_model_id, 
                attn_implementation=config._attn_implementation
            )
            if(new_vocab_size is not None):
                self.language_model.resize_token_embeddings(156940)
        else:
            self.language_model = AutoModelForCausalLM.from_config(
                config.text_config, 
                attn_implementation=config._attn_implementation

            )
        self.pad_token_id = (
            self.config.pad_token_id if self.config.pad_token_id is not None else -1
        )
        self.post_init()

    def get_input_embeddings(self):
        return self.language_model.get_input_embeddings()

    def set_input_embeddings(self, value):
        self.language_model.set_input_embeddings(value)

    def get_output_embeddings(self):
        return self.language_model.get_output_embeddings()

    def set_output_embeddings(self, new_embeddings):
        self.language_model.set_output_embeddings(new_embeddings)

    def set_decoder(self, decoder):
        self.language_model.set_decoder(decoder)

    def get_decoder(self):
        return self.language_model.get_decoder()

    def tie_weights(self):
        return self.language_model.tie_weights()

    def resize_token_embeddings(
        self, new_num_tokens: Optional[int] = None, pad_to_multiple_of=None
    ) -> nn.Embedding:
        model_embeds = self.language_model.resize_token_embeddings(
            new_num_tokens, pad_to_multiple_of
        )
        self.config.text_config.vocab_size = model_embeds.num_embeddings
        self.config.vocab_size = model_embeds.num_embeddings
        self.vocab_size = model_embeds.num_embeddings
        return model_embeds

    def _merge_input_ids_with_audio_features(
        self, audio_features, inputs_embeds, input_ids, attention_mask, labels
    ):
        num_audio_samples, num_audio_patches, embed_dim = audio_features.shape
        batch_size, sequence_length = input_ids.shape
        left_padding = not torch.sum(
            input_ids[:, -1] == torch.tensor(self.pad_token_id)
        )
        special_audio_token_mask = input_ids == self.config.audio_token_index
        num_special_image_tokens = torch.sum(special_audio_token_mask, dim=-1)
        max_embed_dim = (
            num_special_image_tokens.max() * (num_audio_patches - 1)
        ) + sequence_length
        batch_indices, non_audio_indices = torch.where(
            input_ids != self.config.audio_token_index
        )

        new_token_positions = (
            torch.cumsum((special_audio_token_mask * (num_audio_patches - 1) + 1), -1)
            - 1
        )
        nb_image_pad = max_embed_dim - 1 - new_token_positions[:, -1]
        if left_padding:
            new_token_positions += nb_image_pad[:, None]  # offset for left padding
        text_to_overwrite = new_token_positions[batch_indices, non_audio_indices]

        final_embedding = torch.zeros(
            batch_size,
            max_embed_dim,
            embed_dim,
            dtype=inputs_embeds.dtype,
            device=inputs_embeds.device,
        )
        final_attention_mask = torch.zeros(
            batch_size,
            max_embed_dim,
            dtype=attention_mask.dtype,
            device=inputs_embeds.device,
        )
        if labels is not None:
            final_labels = torch.full(
                (batch_size, max_embed_dim),
                self.config.ignore_index,
                dtype=input_ids.dtype,
                device=input_ids.device,
            )

        target_device = inputs_embeds.device
        batch_indices, non_audio_indices, text_to_overwrite = (
            batch_indices.to(target_device),
            non_audio_indices.to(target_device),
            text_to_overwrite.to(target_device),
        )
        attention_mask = attention_mask.to(target_device)


        final_embedding[batch_indices, text_to_overwrite] = inputs_embeds[
            batch_indices, non_audio_indices
        ]
        final_attention_mask[batch_indices, text_to_overwrite] = attention_mask[
            batch_indices, non_audio_indices
        ]
        if labels is not None:
            final_labels[batch_indices, text_to_overwrite] = labels[
                batch_indices, non_audio_indices
            ]

        audio_to_overwrite = torch.all(final_embedding == 0, dim=-1)
        audio_positions = audio_to_overwrite.to(torch.int16).cumsum(-1) - 1
        audio_left_pad_mask = audio_positions >= nb_image_pad[:, None].to(target_device)
        audio_to_overwrite &= audio_left_pad_mask

        if audio_to_overwrite.sum() != audio_features.shape[:-1].numel():
            # print()
            raise ValueError(
                f"The input provided to the model are wrong. The number of audio tokens is {torch.sum(special_audio_token_mask)} while"
                f" the number of audio given to the model is {num_audio_samples}. This prevents correct indexing and breaks batch generation."
            )

        final_embedding[audio_to_overwrite] = (
            audio_features.contiguous()
            .reshape(-1, embed_dim)
            .to(target_device, dtype=final_embedding.dtype)
        )
        final_attention_mask |= audio_to_overwrite
        position_ids = (final_attention_mask.cumsum(-1) - 1).masked_fill_(
            (final_attention_mask == 0), 1
        )

        if labels is None:
            final_labels = None

        return final_embedding, final_attention_mask, final_labels, position_ids

    def get_multimodal_embeddings(self, audio_values):
        speech_embeddings = self.multi_modal_projector(audio_values)
        return speech_embeddings
    
    def forward(
        self,
        input_ids: torch.LongTensor = None,
        audio_values: torch.FloatTensor = None,
        attention_mask: Optional[torch.Tensor] = None,
        position_ids: Optional[torch.LongTensor] = None,
        past_key_values: Optional[List[torch.FloatTensor]] = None,
        inputs_embeds: Optional[torch.FloatTensor] = None,
        labels: Optional[torch.LongTensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        return_dict: Optional[bool] = None,
    ) -> Union[Tuple, OrpheusCausalLMOutputWithPast]:
        

   
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        if inputs_embeds is None:
            inputs_embeds = self.get_input_embeddings()(input_ids)

            if (
                audio_values is not None
                and len(audio_values) > 0
                and audio_values.shape[1] > 1
                and input_ids.shape[1] != 1
            ):
                
                audio_features = self.multi_modal_projector(audio_values)
                (
                    inputs_embeds,
                    attention_mask,
                    labels,
                    position_ids,
                ) = self._merge_input_ids_with_audio_features(
                    audio_features, inputs_embeds, input_ids, attention_mask, labels
                )
                if labels is None:
                    labels = torch.full_like(
                        attention_mask, self.config.ignore_index
                    ).to(torch.long)
            else:

                if (
                    past_key_values is not None
                    and audio_values is not None
                    and input_ids.shape[1] == 1
                ):

                    first_layer_past_key_value = past_key_values[0][0][:, :, :, 0]

                    batch_index, non_attended_tokens = torch.where(
                        first_layer_past_key_value.float().sum(-2) == 0
                    )

                    target_seqlen = first_layer_past_key_value.shape[-1] + 1

                    extended_attention_mask = torch.ones(
                        (
                            attention_mask.shape[0],
                            target_seqlen - attention_mask.shape[1],
                        ),
                        dtype=attention_mask.dtype,
                        device=attention_mask.device,
                    )

                    valid_indices = non_attended_tokens < extended_attention_mask.size(
                        -1
                    )
                    new_batch_index = batch_index[valid_indices]
                    new_non_attended_tokens = non_attended_tokens[valid_indices]

                    extended_attention_mask[
                        new_batch_index, new_non_attended_tokens
                    ] = 0

                    attention_mask = torch.cat(
                        (attention_mask, extended_attention_mask), dim=1
                    )
                    position_ids = torch.sum(attention_mask, dim=1).unsqueeze(-1) - 1

        outputs = self.language_model(
            attention_mask=attention_mask,
            position_ids=position_ids,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )

        logits = outputs[0]

        loss = None
        if labels is not None:
            if attention_mask is not None:
                shift_attention_mask = attention_mask[..., 1:]
                shift_logits = logits[..., :-1, :][
                    shift_attention_mask.to(logits.device) != 0
                ].contiguous()
                shift_labels = labels[..., 1:][
                    shift_attention_mask.to(labels.device) != 0
                ].contiguous()
            else:
                shift_logits = logits[..., :-1, :].contiguous()
                shift_labels = labels[..., 1:].contiguous()
            loss_fct = nn.CrossEntropyLoss()
            loss = loss_fct(
                shift_logits.view(-1, shift_logits.size(-1)),
                shift_labels.view(-1).to(shift_logits.device),
            )

        if not return_dict:
            output = (logits,) + outputs[1:]
            return (loss,) + output if loss is not None else output

        return OrpheusCausalLMOutputWithPast(
            loss=loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )

    def prepare_inputs_for_generation(
        self,
        input_ids,
        past_key_values=None,
        inputs_embeds=None,
        audio_values=None,
        attention_mask=None,
        **kwargs,
    ):
        if past_key_values is not None:
            if isinstance(past_key_values, Cache):
                cache_length = past_key_values.get_seq_length()
                past_length = past_key_values.seen_tokens
            else:
                cache_length = past_length = past_key_values[0][0].shape[2]

            if (
                attention_mask is not None
                and attention_mask.shape[1] > input_ids.shape[1]
            ):
                input_ids = input_ids[:, -(attention_mask.shape[1] - past_length) :]

            elif past_length < input_ids.shape[1]:
                input_ids = input_ids[:, past_length:]
            elif self.config.audio_token_index in input_ids:
                input_ids = input_ids[:, input_ids.shape[1] - 1 :]

            if cache_length < past_length and attention_mask is not None:
                attention_mask = attention_mask[
                    :, -(cache_length + input_ids.shape[1]) :
                ]

        position_ids = kwargs.get("position_ids", None)
        if attention_mask is not None and position_ids is None:
            # create position_ids on the fly for batch generation
            position_ids = attention_mask.long().cumsum(-1) - 1
            position_ids.masked_fill_(attention_mask == 0, 1)
            if past_key_values:
                position_ids = position_ids[:, -input_ids.shape[1] :]

        if inputs_embeds is not None and past_key_values is None:
            model_inputs = {"inputs_embeds": inputs_embeds}
        else:
            model_inputs = {"input_ids": input_ids}

        model_inputs.update(
            {
                "position_ids": position_ids,
                "past_key_values": past_key_values,
                "use_cache": kwargs.get("use_cache"),
                "attention_mask": attention_mask,
                "audio_values": audio_values,
            }
        )
        return model_inputs

    def _reorder_cache(self, *args, **kwargs):
        return self.language_model._reorder_cache(*args, **kwargs)




from huggingface_hub import snapshot_download
from concurrent.futures import ThreadPoolExecutor
import torch
import whisper
from transformers import AutoModel, AutoConfig
from .model import OrpheusConfig, OrpheusForConditionalGeneration
from snac import SNAC


class OrpheusConversation():
    def __init__(self, parent):
        self.existing_embeds = None
        self.current_message = None
        self.model = parent.model
        self.special_tokens = parent.special_tokens
        self.audio_encoder = parent.audio_encoder
        self.tokenizer = parent.tokenizer
        self.parent = parent

    def append_message(self, message):
        self._validate_message(message)
        self.current_message = message

    def _validate_message(self, message):
        if "format" not in message:
            raise ValueError("Message must contain a 'format' key")
        
        if "data" not in message:
            raise ValueError("Message must contain a 'data' key")

        if message["format"] not in ["text", "speech"]:
            raise ValueError("Message format must be either 'text' or 'speech'")
        
        if message["format"] == "speech":
            if type(message["data"]) != torch.Tensor:
                raise ValueError("Speech data must be a torch.Tensor")
            if message["data"].shape[0] != 1:
                raise ValueError("Speech data must be a 1D tensor")
            
        if message["format"] == "text":
            if type(message["data"]) != str:
                raise ValueError("Text data must be a string")

    def _get_embeds(self):
        if self.current_message["format"] == "text":
            return self._get_text_embeds()
        else:
            return self._get_speech_embeds()
        
    def _get_text_embeds(self):
        print("getting text embeds")
        text = self.current_message["data"]
        text_tokens = self.tokenizer(text, return_tensors="pt").input_ids
        text_tokens = text_tokens.to(self.model.device)
        text_embeds = self.model.get_input_embeddings()(text_tokens)

        text_embeds = text_embeds.to(dtype=torch.bfloat16).to(self.model.device)
        start_token = torch.tensor([[128259]], dtype=torch.int64)
        end_tokens = torch.tensor([[128009, 128260]], dtype=torch.int64)

        start_token = start_token.to(self.model.device)
        end_tokens = end_tokens.to(self.model.device)

        start_embeds = self.model.get_input_embeddings()(start_token)
        end_embeds = self.model.get_input_embeddings()(end_tokens)

        start_embeds = start_embeds.to(dtype=torch.bfloat16)
        end_embeds = end_embeds.to(dtype=torch.bfloat16)

        if self.existing_embeds is not None:
            self.existing_embeds = self.existing_embeds.to(self.model.device).to(dtype=torch.bfloat16)
            all_embeds = torch.cat([self.existing_embeds, start_embeds, text_embeds, end_embeds], dim=1)
        else:
            all_embeds = torch.cat([start_embeds, text_embeds, end_embeds], dim=1)
        return all_embeds
    


    
    def _get_speech_embeds(self):
        audio_features = self.parent._get_audio_features(self.current_message["data"])
        audio_features = audio_features.to(dtype=torch.bfloat16).to(self.model.device)
        audio_embeds = self.model.multi_modal_projector(audio_features)
        start_token = torch.tensor([[128259, 128000]], dtype=torch.int64)
        end_tokens = torch.tensor([[128009, 128260]], dtype=torch.int64)
        start_token = start_token.to(self.model.device)
        end_tokens = end_tokens.to(self.model.device)
        start_embeds = self.model.get_input_embeddings()(start_token)
        end_embeds = self.model.get_input_embeddings()(end_tokens)
        start_embeds = start_embeds.to(dtype=torch.bfloat16)
        end_embeds = end_embeds.to(dtype=torch.bfloat16)
        if self.existing_embeds is not None:
            all_embeds = torch.cat([self.existing_embeds, start_embeds, audio_embeds, end_embeds], dim=1)
        else:
            all_embeds = torch.cat([start_embeds, audio_embeds, end_embeds], dim=1)

        return all_embeds
    
    def _update_existing_embeds(self, output_tokens):
        output_embeddings = self.model.get_input_embeddings()(output_tokens)
        end_of_ai_embedding = self.model.get_input_embeddings()(torch.tensor([[self.special_tokens["end_of_ai"]]]).to(self.model.device))
        if self.existing_embeds is None:
            self.existing_embeds = torch.cat([output_embeddings, end_of_ai_embedding], dim=1).to("cpu")
        else:
            all_embeddings = torch.cat([self.existing_embeds, output_embeddings, end_of_ai_embedding], dim=1).to("cpu")
            self.existing_embeds = all_embeddings

    
    def generate_response(self):
        if self.current_message is None:
            raise ValueError("Please append a message first")
        
        embeds = self._get_embeds()
 
        output_tokens = self.model.generate(
            inputs_embeds=embeds, 
            max_new_tokens=5000, 
            temperature=0.9,
            repetition_penalty=1.2, 
            eos_token_id=self.special_tokens["end_of_speech"],
            )
        
        self._update_existing_embeds(output_tokens)

        print("successffully updated existing embeds")

        print(output_tokens.shape)

        output = self.parent.parse_output_tokens(output_tokens)
        return output
        



class OrpheusUtility():
    def __init__(self,
                 text_model_name="amuvarma/3b-zuckreg-convo",
                 multimodal_model_name="amuvarma/zuck-3bregconvo-automodelcompat"
                 ):
        self.special_tokens = {
            "start_of_text": 128000,
            "end_of_text": 128009,
            "start_of_speech": 128257,
            "end_of_speech": 128258,
            "start_of_human": 128259,
            "end_of_human": 128260,
            "start_of_ai": 128261,
            "end_of_ai": 128262,
            "pad_token": 128263
        }
        self._is_model_initialised = False
        self._is_model_downloaded = False
        self.audio_encoder = whisper.load_model("small")

        snac_model = SNAC.from_pretrained("hubertsiuzdak/snac_24khz")
        snac_model = snac_model.to("cuda")
        self.snac_model = snac_model

        pass

    def _verify_initialisation(self):
        if not self._is_model_initialised:
            raise ValueError("Please ensure you have registered the model with the OrpheusUtility class using orpheus.register_auto_model(model)")
        return self._is_model_initialised

    def _download_from_hub(self, model_name):
        snapshot_download(
            repo_id=model_name,
            allow_patterns=[
                "config.json",
                "*.safetensors",
                "model.safetensors.index.json",
            ],
            ignore_patterns=[
                "optimizer.pt",
                "pytorch_model.bin",
                "training_args.bin",
                "scheduler.pt",
                "tokenizer.json",
                "tokenizer_config.json",
                "special_tokens_map.json",
                "vocab.json",
                "merges.txt",
                "tokenizer.*"
            ]
        )

    def initialise(self, text_model_name="amuvarma/3b-zuckreg-convo", multimodal_model_name="amuvarma/zuck-3bregconvo-automodelcompat"):
        with ThreadPoolExecutor(max_workers=2) as executor:
            future_text = executor.submit(
                self._download_from_hub, text_model_name)
            future_multimodal = executor.submit(
                self._download_from_hub, multimodal_model_name)
            future_text.result()
            future_multimodal.result()
        self._is_model_downloaded = True

        AutoConfig.register("orpheus", OrpheusConfig)
        AutoModel.register(OrpheusConfig, OrpheusForConditionalGeneration)


        print("Downloads complete!")

    def register_auto_model(self, model, tokenizer):

        assert model is not None, "You must provide a model"
        assert tokenizer is not None, "You must provide a tokenizer"

        self.model = model
        self.tokenizer = tokenizer
        self._is_model_initialised = True


    def _get_input_from_text(self, text):
        input_ids = self.tokenizer.encode(text, return_tensors="pt")
        input_ids = torch.cat([
            torch.tensor([[self.special_tokens["start_of_human"]]]),
            input_ids, 
            torch.tensor([[self.special_tokens["end_of_text"], 
            self.special_tokens["end_of_human"]]])], 
        dim=1)
        input_ids = input_ids.to("cuda")
        return {"input_ids": input_ids}
    

    def _process_audio_tensor(self, audio, sample_rate=16000):
        audio = audio.to(torch.float32)
        duration_ms = (len(audio) / sample_rate) * 1000
        audio = whisper.pad_or_trim(audio)
        mel = whisper.log_mel_spectrogram(audio)
        return mel, int(duration_ms / 20) + 1
    
    def _get_audio_features(self, speech):
        audio_input = speech.squeeze(0)
        mel, length = self._process_audio_tensor(audio_input)
        mel = mel.to("cuda")
        mel = mel.unsqueeze(0)
        audio_feature = self.audio_encoder.embed_audio(mel)[0][:length]
        audio_feature = audio_feature.unsqueeze(0)
        return audio_feature
    
    def _get_input_from_speech(self, speech):
        self._verify_initialisation()
        audio_features = self._get_audio_features(speech)
        audio_features = audio_features.to(dtype=torch.bfloat16).to("cuda")
        audio_embeds = self.model.multi_modal_projector(audio_features)
        start_token = torch.tensor([[self.special_tokens["start_of_human"], self.special_tokens["start_of_text"]]], dtype=torch.int64)
        end_tokens = torch.tensor([[self.special_tokens["end_of_text"], self.special_tokens["end_of_human"]]], dtype=torch.int64)
        start_token = start_token.to("cuda")
        end_tokens = end_tokens.to("cuda")
        start_embeds = self.model.get_input_embeddings()(start_token)
        end_embeds = self.model.get_input_embeddings()(end_tokens)
        start_embeds = start_embeds.to(dtype=torch.bfloat16)
        end_embeds = end_embeds.to(dtype=torch.bfloat16)
        all_embeds = torch.cat([start_embeds, audio_embeds, end_embeds], dim=1)
        return {"inputs_embeds": all_embeds}
    
    def get_inputs(self, text=None, speech=None):
        if text is None and speech is None:
            raise ValueError("Either text or speech must be provided")
        if text is not None and speech is not None:
            raise ValueError("Only one of text or speech must be provided")
        

        if text is not None:
            return self._get_input_from_text(text)
        else:
            return self._get_input_from_speech(speech)
        
    
    def _redistribute_codes(self, code_list):
        layer_1 = []
        layer_2 = []
        layer_3 = []
        for i in range((len(code_list)+1)//7):
            layer_1.append(code_list[7*i])
            layer_2.append(code_list[7*i+1]-4096)
            layer_3.append(code_list[7*i+2]-(2*4096))
            layer_3.append(code_list[7*i+3]-(3*4096))
            layer_2.append(code_list[7*i+4]-(4*4096))
            layer_3.append(code_list[7*i+5]-(5*4096))
            layer_3.append(code_list[7*i+6]-(6*4096))

        codes = [torch.tensor(layer_1).unsqueeze(0).to("cuda"),
                torch.tensor(layer_2).unsqueeze(0).to("cuda"),
                torch.tensor(layer_3).unsqueeze(0).to("cuda")]
        audio_hat = self.snac_model.decode(codes)
        return audio_hat
    
    def _get_waveform_from_tokens (self, output_tokens):

        token_to_find = self.special_tokens["start_of_speech"]
        token_to_remove = self.special_tokens["pad_token"]

        if token_to_find not in output_tokens:
            return None
        
        token_indices = (output_tokens == token_to_find).nonzero(as_tuple=True)

        if len(token_indices[-1]) > 0:
            last_occurrence_idx = token_indices[1][-1].item()
            cropped_tensor = output_tokens[:, last_occurrence_idx+1:]
        else:
            cropped_tensor = output_tokens

        mask = cropped_tensor != token_to_remove
        cropped_tensor = cropped_tensor[mask].view(cropped_tensor.size(0), -1)

        processed_tensor = cropped_tensor - 128266
        original_shape = processed_tensor.shape
        new_dim_1 = (original_shape[1] // 7) * 7
        processed_tensor = processed_tensor[:, :new_dim_1]
        code_list = processed_tensor[0].tolist()
        samples = self._redistribute_codes(code_list)

        waveform = samples.detach().squeeze().to("cpu").numpy()
        return waveform
        
    def _get_text_from_tokens(self, output_tokens):
        
        token_to_find = self.special_tokens["start_of_ai"]
        end_token = self.special_tokens["end_of_text"]

        print("getting text")

        if token_to_find not in output_tokens:
            return None

        token_indices = (output_tokens == token_to_find).nonzero(as_tuple=True)
        
        print("token indices", token_indices)

        if len(token_indices[1]) > 0:

            start_idx = token_indices[1][-1].item() + 1
            end_indices = (output_tokens[:, start_idx:] == end_token).nonzero(as_tuple=True)
            
            print("end indices", end_indices)
            if len(end_indices[1]) > 0:
                end_idx = start_idx + end_indices[1][0].item()
                text_tokens = output_tokens[:, start_idx:end_idx]
                decoded_text = self.tokenizer.decode(text_tokens[0])
                print("decoded text", decoded_text)
                return decoded_text
                
        return None


    def parse_output_tokens (self, output_tokens):
        waveform = self._get_waveform_from_tokens(output_tokens)
        print(waveform)
        text = self._get_text_from_tokens(output_tokens)
        response_dict = {"waveform": waveform, "text": text}
        if waveform is None:
            response_dict["waveform"] = "No tokens generated to output speech, please increase number of tokens generated"
        if text is None:
            response_dict["text"] = "No tokens generated to output text. There may be an error, or the number of tokens this model is set to generate is too low."
        return response_dict


    def initialise_conversation_model(self):
        return OrpheusConversation(self)



import os.path

# Get the directory containing this __init__.py file
ASSETS_DIR = os.path.dirname(os.path.abspath(__file__))

# Define the path to speech.wav
SPEECH_WAV_PATH = os.path.join(ASSETS_DIR, 'speech.wav')

# Make these available when importing from assets
__all__ = ['SPEECH_WAV_PATH', 'ASSETS_DIR']


from transformers import PretrainedConfig, CONFIG_MAPPING

class OrpheusConfig(PretrainedConfig):
    
    model_type = "orpheus"

    is_composition = False
    def __init__(
        self,
        text_config=None,
        audio_hidden_size = 768,
        ignore_index=-100,
        audio_token_index=32000,
        vocab_size=32000,
        hidden_size=3072,
        stack_factor=8,
        projector_type="mlp",
        
        
        **kwargs,
    ):
        self.ignore_index = ignore_index
        self.audio_token_index = audio_token_index
        self.vocab_size = vocab_size
        self.audio_hidden_size = audio_hidden_size
        self.text_config = text_config

        self.text_model_id = None

        self.hidden_size = hidden_size
        self.stack_factor = stack_factor
        self.projector_type = projector_type

        if isinstance(self.text_config, dict):
            text_config["model_type"] = (
                text_config["model_type"] if "model_type" in text_config else "llama"
            )
            self.text_config = CONFIG_MAPPING[text_config["model_type"]](**text_config)
            self.vocab_size = self.text_config.vocab_size
        elif text_config is None:
            self.text_config = CONFIG_MAPPING["llama"]()
        
        super().__init__(**kwargs)
